
//
//  Bench.c - Packed Boolean Logic
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 26 Jan, 2002 - James Leiterman
//
//  Last Modification
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries
    
#include <stdio.h>
#include <assert.h>

    //  Book Development Library Modules
    
#include "CpuAsm.h"                                     // CPU Module
#include "vmp3D.h"
#include "PBool.h"
#include "PBoolEx.h"



typedef enum
{
    iTEST_PAND,             // Packed AND
    iTEST_PANDC,            // Packed AND w/1's complement
    iTEST_POR,              // Packed OR
    iTEST_PNOR,             // Packed 1's complement of OR
    iTEST_PXOR,             // Packed Exclusive OR
    iTEST_PNOT              // Packed 1's complement
} iTEST;




//
//  Compare two 128bit bit blocks
//

bool vmp_IsPVEqual( void * const pvA, void * const pvB )
{
    ASSERT_PTR4(pvA);
    ASSERT_PTR4(pvB);
    
    if ( memcmp( pvA, pvB, 16 ) )
      {
        return false;
      }

    return true;
}                                               



//
//  Test Packed Shift 128bit
//
static void test_pV( bool bAlign, iTEST iFunc )
{
    uint    count;
    int8    v, buf[4*16 +15], *bt, *bu, *bv, *bw;

    if (bAlign)
      {
        bt = (int8*) ALIGN16((int)buf);
      }
    else
      {
        bt = buf;
        if (0== (((uint)bt)&0x3)) bt += 4;
      }

    bu = bt + (128/8);
    bv = bu + (128/8);
    bw = bv + (128/8);

    v = 0;
    do {
        for ( count=0; count < 8+1; count++ )
          {
            bt[0]=v;     bt[1]=v+1;   bt[2]=v+2;   bt[3]=v+3;
            bt[4]=v+4;   bt[5]=v+5;   bt[6]=v+6;   bt[7]=v+7;
            bt[8]=v+8;   bt[9]=v+9;   bt[10]=v+10; bt[11]=v+11;
            bt[12]=v+12; bt[13]=v+13; bt[14]=v+14; bt[15]=v+15;

            bw[0]=v-1;   bw[1]=v-3;   bw[2]=v-5;   bw[3]=v-7;
            bw[4]=v-9;   bw[5]=v-11;  bw[6]=v-13;  bw[7]=v-15;
            bw[8]=v-17;  bw[9]=v-19;  bw[10]=v-23; bw[11]=v-29;
            bw[12]=v-31; bw[13]=v-37; bw[14]=v-41; bw[15]=v-43;

            switch (iFunc)
              {
                case iTEST_PAND:                        // AND
                    vmp_pandGeneric(bu, bt, bw);
                    if ( bAlign )
                        vmp_pand(   bv, bt, bw);
                    else        
                        vmp_pand0(  bv, bt, bw);
                    break;

                case iTEST_PANDC:                       // AND 1's complement
                    vmp_pandcGeneric( bu, bt, bw);
                    if ( bAlign )
                        vmp_pandc(    bv, bt, bw);
                    else         
                        vmp_pandc0(   bv, bt, bw);
                    break;

                case iTEST_POR:                         // OR
                    vmp_porGeneric(   bu, bt, bw);
                    if ( bAlign )
                        vmp_por(      bv, bt, bw);
                    else
                        vmp_por0(     bv, bt, bw);
                    break;

                case iTEST_PNOR:                        // NOR
                    vmp_pnorGeneric(bu, bt, bw);
                    if ( bAlign )
                        vmp_pnor(   bv, bt, bw);
                    else
                        vmp_pnor0(  bv, bt, bw);
                    break;

                case iTEST_PXOR:                        // XOR
                    vmp_pxorGeneric(bu, bt, bw);
                    if ( bAlign )
                        vmp_pxor(   bv, bt, bw);
                    else
                        vmp_pxor0(  bv, bt, bw);
                    break;

                case iTEST_PNOT:                        // NOT
                    vmp_pnotGeneric(bu, bt);
                    if ( bAlign )
                        vmp_pnot(   bv, bt);
                    else
                        vmp_pnot0(  bv, bt);
                    break;
              }

            if ( !vmp_IsPVEqual( bu, bv ) )
              {
                printf( "Mismatch on %u and count %u\n", v, count );
              }
          }
      } while( 0 != ++v);
}




int main( int argc, char *argv[] )
{
    CpuInfo cinfo;
    char szBuf[ CPU_SZBUF_MAX ];

    CpuDetect( &cinfo );                                // Detect CPU
    
    printf( "\nPacked (Vector) Boolean Logic Test\n\n" );
    printf( "%s\n", CpuInfoStr( szBuf, &cinfo ));
    
    PBoolSetup( &cinfo );                               // Setup our Vector Boolean stuff

    printf( "Testing (Vector) AND\n" );
    test_pV( false, iTEST_PAND );

    printf( "Testing (Vector) AND (Aligned)\n" );
    test_pV( true,  iTEST_PAND );

    printf( "Testing (Vector) AND w/1's complement\n" );
    test_pV( false, iTEST_PANDC );

    printf( "Testing (Vector) AND w/1's complement (Aligned)\n" );
    test_pV( true,  iTEST_PANDC );

    printf( "Testing (Vector) OR\n" );
    test_pV( false, iTEST_POR );

    printf( "Testing (Vector) OR (Aligned)\n" );
    test_pV( true,  iTEST_POR );

    printf( "Testing (Vector) 1's complement OR (NOR)\n" );
    test_pV( false, iTEST_PNOR );

    printf( "Testing (Vector) 1's complement OR (NOR) (Aligned)\n" );
    test_pV( true,  iTEST_PNOR );

    printf( "Testing (Vector) XOR\n" );
    test_pV( false, iTEST_PXOR );

    printf( "Testing (Vector) XOR (Aligned)\n" );
    test_pV( true,  iTEST_PXOR );

    printf( "Testing (Vector) 1's complement (NOT)\n" );
    test_pV( false, iTEST_PNOT );

    printf( "Testing (Vector) 1's complement (NOT) (Aligned)\n" );
    test_pV( true,  iTEST_PNOT );

    printf( "Fin!\n\n" );

    return 0;
}

