
// 
//  QVas3DVU.c - 3D (Quad) Vector Addition & Subtraction
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 3 Oct 2001 - James Leiterman
//
//  Last Modification
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries
    
#include    <assert.h>
#include    <stdio.h>

#ifdef CC_VMP_PS2
#ifndef HOMEBREW
#include    <libvu0.h>
#endif
#endif

    //  Book Development Library Modules

#include    "CpuAsm.h"                                  // CPU
#include    "vmp3D.h"
#include    "vmp3DEx.h"
#include    "QVas3D.h"                                  // 3D Math
#include    "QVas3DEx.h"



#ifdef HOMEBREW
typedef float       sceVu0FVECTOR[4] __attribute__((aligned (16)));
#define USE_ASM
#endif



//
//  (Quad) Vector Summation (Addition) - Unaligned
//

void vmp_QVecAddVU0(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB)
{
    sceVu0FVECTOR vD, vA, vB;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);
    ASSERT_PTR4(pvB);

    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
    vA[3] = pvA->w;
        
    vB[0] = pvB->x;
    vB[1] = pvB->y;
    vB[2] = pvB->z;
    vB[3] = pvB->w;

#ifdef USE_ASM
    asm __volatile__("

    lqc2    vf4,0x0(%1)
    lqc2    vf5,0x0(%2)
    
    vadd.xyzw   vf6,vf4,vf5     # D=A+B

    sqc2    vf6,0x0(%0)

    ":
     : "r" (vD) , "r" (vA), "r" (vB) 
    );

#else
    sceVu0AddVector(vD, vA, vB);
#endif

    pvD->x = vD[0];
    pvD->y = vD[1];
    pvD->z = vD[2];
    pvD->w = vD[3];
}



//
//  (Quad) Vector Summation (Addition) - Aligned
//

void vmp_QVecAddVU(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB)
{
    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);
    ASSERT_PTR16(pvB);

#ifdef USE_ASM
    asm __volatile__("

    lqc2    vf4,0x0(%1)
    lqc2    vf5,0x0(%2)
    
    vadd.xyzw   vf6,vf4,vf5     # D.xyzw = A.xyzw + B.xyzw

    sqc2    vf6,0x0(%0)

    ":
     : "r" (pvD) , "r" (pvA), "r" (pvB) 
    );

#else
    sceVu0AddVector(*(sceVu0FVECTOR *)pvD, *(sceVu0FVECTOR *)pvA, *(sceVu0FVECTOR *)pvB);
#endif
}



//______________________________________________________
//
//  (Quad) Vector Difference (Subtraction) - Unaligned
//

void vmp_QVecSubVU0(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB)
{
    sceVu0FVECTOR vD, vA, vB;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);
    ASSERT_PTR4(pvB);

    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
    vA[3] = pvA->w;
        
    vB[0] = pvB->x;
    vB[1] = pvB->y;
    vB[2] = pvB->z;
    vB[3] = pvB->w;

#ifdef USE_ASM
    asm __volatile__("

    lqc2    vf4,0x0(%1)
    lqc2    vf5,0x0(%2)
    
    vsub.xyzw   vf6,vf4,vf5     # D.xyzw = A.xyzw - B.xyzw

    sqc2    vf6,0x0(%0)

    ":
     : "r" (vD) , "r" (vA), "r" (vB) 
    );

#else
    sceVu0SubVector(vD, vA, vB);
#endif

    pvD->x = vD[0];
    pvD->y = vD[1];
    pvD->z = vD[2];
    pvD->w = vD[3];
}



//
//  (Quad) Vector Difference (Subtraction) - Aligned
//

void vmp_QVecSubVU(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB)
{
    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);
    ASSERT_PTR16(pvB);

#ifdef USE_ASM
    asm __volatile__("

    lqc2    vf4,0x0(%1)
    lqc2    vf5,0x0(%2)
    
    vsub.xyzw   vf6,vf4,vf5     # D.xyzw = A.xyzw - B.xyzw

    sqc2    vf6,0x0(%0)

    ":
     : "r" (pvD) , "r" (pvA), "r" (pvB) 
    );

#else
    sceVu0SubVector(*(sceVu0FVECTOR *)pvD, *(sceVu0FVECTOR *)pvA, *(sceVu0FVECTOR *)pvB);
#endif
}



//______________________________________________________
//
//  (Quad) Vector Scalar Summation (Addition) - Unaligned
//

void vmp_QVecAddScalarVU0(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, float fScalar)
{
#ifdef USE_ASM
    sceVu0FVECTOR vD, vA;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);

    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
    vA[3] = pvA->w;
    
    asm __volatile__("

    lqc2    vf4,0x0(%1)
    mfc1    $8,%2               #fA
    qmtc2    $8,vf5             #vf5=  xfer FP from EE to VU
    
    vaddx.xyzw  vf6, vf4, vf5   # D.xyzw =A.xyzw + B.x

    sqc2    vf6,0x0(%0)

    ":
     : "r" (vD) , "r" (vA), "f" (fScalar) 
     : "$8"
    );

#else
    sceVu0FVECTOR vD, vA, vB;

    ASSERT_PTR(pvD);
    ASSERT_PTR(pvA);

    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
    vA[3] = pvA->w;
    
    vB[0] = vB[1] = vB[2] = vB[3] = fScalar;

    sceVu0AddVector(vD, vA, vB);
#endif

    pvD->x = vD[0];
    pvD->y = vD[1];
    pvD->z = vD[2];
    pvD->w = vD[3];
}



//
//  (Quad) Vector Scalar Summation (Addition) - Aligned
//

void vmp_QVecAddScalarVU(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, float fScalar)
{
#ifdef USE_ASM
    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);

    asm __volatile__("

    lqc2    vf4,0x0(%1)
    mfc1    $8,%2               #fA
    qmtc2    $8,vf5             #vf5=  xfer FP from EE to VU
    
    vaddx.xyzw  vf6, vf4, vf5   # D.xyzw = A.xyzw + B.x

    sqc2    vf6,0x0(%0)

    ":
     : "r" (pvD) , "r" (pvA), "f" (fScalar) 
     : "$8"
    );

#else
    sceVu0FVECTOR vB;

    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);

    vB[0] = vB[1] = vB[2] = vB[3] = fScalar;

    sceVu0AddVector(*(sceVu0FVECTOR *)pvD, *(sceVu0FVECTOR *)pvA, vB);
#endif
}


//______________________________________________________
//
//  (Quad) Vector Scalar Difference (Subtraction) - Unaligned
//

void vmp_QVecSubScalarVU0(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, float fScalar)
{   
#ifdef USE_ASM
    sceVu0FVECTOR vD, vA;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);

    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
    vA[3] = pvA->w;
    
    asm __volatile__("

    lqc2    vf4,0x0(%1)
    mfc1    $8,%2               #fA
    qmtc2    $8,vf5             #vf5=  xfer FP from EE to VU
    
    vsubx.xyzw  vf6, vf4, vf5   # D.xyzw = A.xyzw - B.x

    sqc2    vf6,0x0(%0)

    ":
     : "r" (vD) , "r" (vA), "f" (fScalar) 
     : "$8"
    );

#else
    sceVu0FVECTOR vD, vA, vB;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);

    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
    vA[3] = pvA->w;
        
    vB[0] = vB[1] = vB[2] = vB[3] = fScalar;

    sceVu0SubVector(vD, vA, vB);
#endif

    pvD->x = vD[0];
    pvD->y = vD[1];
    pvD->z = vD[2];
    pvD->w = vD[3];
}



//
//  (Quad) Vector Scalar Difference (Subtraction) - Aligned
//

void vmp_QVecSubScalarVU(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, float fScalar)
{   
#ifdef USE_ASM
    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);

    asm __volatile__("

    lqc2    vf4,0x0(%1)
    mfc1    $8,%2               #fA
    qmtc2    $8,vf5             #vf5=  xfer FP from EE to VU
    
    vsubx.xyzw  vf6, vf4, vf5   # D.xyzw = A.xyzw - B.x

    sqc2    vf6,0x0(%0)

    ":
     : "r" (pvD) , "r" (pvA), "f" (fScalar) 
     : "$8"
    );

#else
    sceVu0FVECTOR vB;

    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);

    vB[0] = vB[1] = vB[2] = vB[3] = fScalar;

    sceVu0SubVector(*(sceVu0FVECTOR *)pvD, *(sceVu0FVECTOR *)pvA, vB);
#endif
}



//______________________________________________________
//
//  Setup Vec3D module
//
//  NOTE: This function sets up the function vectors so the best function
//      is attached to the vector for the processor running.
//

void QVas3DSetup( const CpuInfo * const pcinfo )
{
    ASSERT_PTR4(pcinfo);
    
    if ( CPUBITS_VU0 & pcinfo->nBits )
      {
        vmp_QVecAdd =           vmp_QVecAddVU;
        vmp_QVecAdd0 =          vmp_QVecAddVU0;
        vmp_QVecSub =           vmp_QVecSubVU;
        vmp_QVecSub0 =          vmp_QVecSubVU0;
        vmp_QVecAddScalar =     vmp_QVecAddScalarVU;
        vmp_QVecAddScalar0 =    vmp_QVecAddScalarVU0;
        vmp_QVecSubScalar =     vmp_QVecSubScalarVU;
        vmp_QVecSubScalar0 =    vmp_QVecSubScalarVU0;
      }

    else
      {
        vmp_QVecAdd =           vmp_QVecAddGeneric;
        vmp_QVecAdd0 =          vmp_QVecAddGeneric;
        vmp_QVecSub =           vmp_QVecSubGeneric;
        vmp_QVecSub0 =          vmp_QVecSubGeneric;
        vmp_QVecAddScalar =     vmp_QVecAddScalarGeneric;
        vmp_QVecAddScalar0 =    vmp_QVecAddScalarGeneric;
        vmp_QVecSubScalar =     vmp_QVecSubScalarGeneric;
        vmp_QVecSubScalar0 =    vmp_QVecSubScalarGeneric;
      }
}




