
//
//  Vas3DVU.c - 3D Vector Addition & Subtraction
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  NOTE: Because these a 3 float Vectors the 4th float has to be handled as 
//      a special case thus slowing the code down. This is the same as an AoS 
//      orientation. Instead a SoA should be used in conjunction with the quad
//      set of functions.
//
//  Created - 3 Oct 2001 - James Leiterman
//
//  Last Modification
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries
    
#include    <assert.h>
#include    <stdio.h>

#ifdef CC_VMP_PS2
#ifndef HOMEBREW
#include    <libvu0.h>
#endif
#endif

    //  Book Development Library Modules

#include    "CpuAsm.h"                                  // CPU
#include    "vmp3D.h"
#include    "vmp3DEx.h"
#include    "Vas3D.h"                                   // 3D Math
#include    "Vas3DEx.h"



#ifdef HOMEBREW
typedef float       sceVu0FVECTOR[4] __attribute__((aligned (16)));
#define USE_ASM
#endif



//
//  Vector Summation (Addition) - Unaligned
//

void vmp_VecAddVU0(vmp3DVector * const pvD, const vmp3DVector * const pvA, const vmp3DVector * const pvB)
{
    sceVu0FVECTOR vD, vA, vB;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);
    ASSERT_PTR4(pvB);

    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
        
    vB[0] = pvB->x;
    vB[1] = pvB->y;
    vB[2] = pvB->z;

#ifdef USE_ASM
    asm __volatile__("

    lqc2    vf4,0x0(%1)
    lqc2    vf5,0x0(%2)
    
    vadd.xyz    vf6,vf4,vf5         # D.w |= A.xyz + B.xyz

    sqc2    vf6,0x0(%0)

    ":
     : "r" (vD) , "r" (vA), "r" (vB) 
    );

#else
    sceVu0AddVector(vD, vA, vB);
#endif

    pvD->x = vD[0];
    pvD->y = vD[1];
    pvD->z = vD[2];
}


//
//  Vector Summation (Addition) - Aligned
//  NOTE: Only difference is assertion!
//

void vmp_VecAddVU(vmp3DVector * const pvD, const vmp3DVector * const pvA, const vmp3DVector * const pvB)
{
#ifdef USE_ASM
    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);
    ASSERT_PTR16(pvB);

    asm __volatile__("

    lqc2    vf4,0x0(%1)
    lqc2    vf5,0x0(%2)
    lqc2    vf6,0x0(%0)             # .xyzw
        
    vadd.xyz    vf6,vf4,vf5         # D.w |= A.xyz + B.xyz

    sqc2    vf6,0x0(%0)             # .xyzw

    ":
     : "r" (pvD) , "r" (pvA), "r" (pvB) 
    );

#else
    sceVu0FVECTOR vD;

    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);
    ASSERT_PTR16(pvB);

    sceVu0AddVector(vD, *(sceVu0FVECTOR *)pvA, *(sceVu0FVECTOR *)pvB);

    sceVu0CopyVectorXYZ(*(sceVu0FVECTOR *)pvD, vD );
#endif
}



//______________________________________________________
//
//  Vector Difference (Subtraction) - Unaligned
//

void vmp_VecSubVU0(vmp3DVector * const pvD, const vmp3DVector * const pvA, const vmp3DVector * const pvB)
{
    sceVu0FVECTOR vD, vA, vB;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);
    ASSERT_PTR4(pvB);

    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
        
    vB[0] = pvB->x;
    vB[1] = pvB->y;
    vB[2] = pvB->z;

#ifdef USE_ASM
    asm __volatile__("

    lqc2    vf4,0x0(%1)
    lqc2    vf5,0x0(%2)
    
    vsub.xyz    vf6,vf4,vf5         # D.w |= A.xyz - B.xyz

    sqc2    vf6,0x0(%0)

    ":
     : "r" (vD) , "r" (vA), "r" (vB) 
    );

#else
    sceVu0SubVector(vD, vA, vB);
#endif

    pvD->x = vD[0];
    pvD->y = vD[1];
    pvD->z = vD[2];
}


//
//  Vector Difference (Subtraction) - Aligned
//  NOTE: Only difference is assertion!
//

void vmp_VecSubVU(vmp3DVector * const pvD, const vmp3DVector * const pvA, const vmp3DVector * const pvB)
{
#ifdef USE_ASM
    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);
    ASSERT_PTR16(pvB);

    asm __volatile__("

    lqc2    vf4,0x0(%1)
    lqc2    vf5,0x0(%2)
    lqc2    vf6,0x0(%0)             # .xyzw
        
    vsub.xyz    vf6,vf4,vf5         # D.w |= A.xyz - B.xyz

    sqc2    vf6,0x0(%0)             # .xyzw

    ":
     : "r" (pvD) , "r" (pvA), "r" (pvB) 
    );

#else
    sceVu0FVECTOR vD;

    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);
    ASSERT_PTR16(pvB);

    sceVu0SubVector(vD, *(sceVu0FVECTOR *)pvA, *(sceVu0FVECTOR *)pvB);

    sceVu0CopyVectorXYZ(*(sceVu0FVECTOR *)pvD, vD );
#endif
}


//______________________________________________________
//
//  Vector Scalar Summation (Addition) - Unaligned
//

void vmp_VecAddScalarVU0(vmp3DVector * const pvD, const vmp3DVector * const pvA, float fScalar)
{
#ifdef USE_ASM
    sceVu0FVECTOR vD, vA;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);

    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
    
    asm __volatile__("

    lqc2    vf4,0x0(%1)
    mfc1    $8,%2               #fA
    qmtc2    $8,vf5             #vf5=  xfer FP from EE to VU
    
    vaddx.xyz   vf6, vf4, vf5   # D.w |= A.xyz + B.x

    sqc2    vf6,0x0(%0)

    ":
     : "r" (vD) , "r" (vA), "f" (fScalar) 
     : "$8"
    );

#else
    sceVu0FVECTOR vD, vA, vB;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);

    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
    vA[3] = 0.0f;
        
    vB[0] = vB[1] = vB[2] = fScalar;
    vB[3] = 0.0f;

    sceVu0AddVector(vD, vA, vB);
#endif

    pvD->x = vD[0];
    pvD->y = vD[1];
    pvD->z = vD[2];
}


//
//  Vector Scalar Summation (Addition) - Aligned
//  NOTE: Only difference is assertion!
//

void vmp_VecAddScalarVU(vmp3DVector * const pvD, const vmp3DVector * const pvA, float fScalar)
{
#ifdef USE_ASM
    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);

    asm __volatile__("

    lqc2    vf4,0x0(%1)
    mfc1    $8,%2               #fA
    lqc2    vf6,0x0(%0)
    qmtc2    $8,vf5             #vf5=  xfer FP from EE to VU
    
    vaddx.xyz   vf6, vf4, vf5   # D.w |= A.xyz + B.x

    sqc2    vf6,0x0(%0)

    ":
     : "r" (pvD) , "r" (pvA), "f" (fScalar) 
     : "$8"
    );

#else
    sceVu0FVECTOR vD, vB;

    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);

    vB[0] = vB[1] = vB[2] = fScalar;
    vB[3] = 0.0f;

    sceVu0AddVector(vD, *(sceVu0FVECTOR *)pvA, vB);

    sceVu0CopyVectorXYZ(*(sceVu0FVECTOR *)pvD, vD );
#endif
}



//______________________________________________________
//
//  Vector Scalar Difference (Subtraction) - Unaligned
//

void vmp_VecSubScalarVU0(vmp3DVector * const pvD, const vmp3DVector * const pvA, float fScalar)
{   
#ifdef USE_ASM
    sceVu0FVECTOR vD, vA;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);

    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
    
    asm __volatile__("

    lqc2    vf4,0x0(%1)
    mfc1    $8,%2               #fA
    qmtc2    $8,vf5             #vf5=  xfer FP from EE to VU
    
    vsubx.xyz   vf6, vf4, vf5   # D.w |= A.xyz - B.x

    sqc2    vf6,0x0(%0)

    ":
     : "r" (vD) , "r" (vA), "f" (fScalar) 
     : "$8"
    );

#else
    sceVu0FVECTOR vD, vA, vB;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);

    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
    vA[3] = 0.0f;
        
    vB[0] = vB[1] = vB[2] = fScalar;
    vB[3] = 0.0f;

    sceVu0SubVector(vD, vA, vB);
#endif

    pvD->x = vD[0];
    pvD->y = vD[1];
    pvD->z = vD[2];
}




//
//  Vector Scalar Difference (Subtraction) - Aligned
//  NOTE: Only difference is assertion!
//

void vmp_VecSubScalarVU(vmp3DVector * const pvD, const vmp3DVector * const pvA, float fScalar)
{   
#ifdef USE_ASM
    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);

    asm __volatile__("

    lqc2    vf4,0x0(%1)
    mfc1    $8,%2               #fA
    lqc2    vf6,0x0(%0)
    qmtc2    $8,vf5             #vf5=  xfer FP from EE to VU
    
    vsubx.xyz   vf6, vf4, vf5   # D.w |= A.xyz - B.x

    sqc2    vf6,0x0(%0)

    ":
     : "r" (pvD) , "r" (pvA), "f" (fScalar) 
     : "$8"
    );

#else
    sceVu0FVECTOR vD, vB;

    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);

    vB[0] = vB[1] = vB[2] = fScalar;
    vB[3] = 0.0f;

    sceVu0SubVector(vD, *(sceVu0FVECTOR *)pvA, vB);

    sceVu0CopyVectorXYZ(*(sceVu0FVECTOR *)pvD, vD );
#endif
}




//______________________________________________________
//
//  Setup Vec3D module
//
//  NOTE: This function sets up the function vectors so the best function
//      is attached to the vector for the processor running.
//

void Vas3DSetup( const CpuInfo * const pcinfo )
{
    ASSERT_PTR4(pcinfo);

    if ( CPUBITS_VU0 & pcinfo->nBits )
      {
        vmp_VecAdd =            vmp_VecAddVU;
        vmp_VecAdd0 =           vmp_VecAddVU0;
        vmp_VecSub =            vmp_VecSubVU;
        vmp_VecSub0 =           vmp_VecSubVU0;
        vmp_VecAddScalar =      vmp_VecAddScalarVU;
        vmp_VecAddScalar0 =     vmp_VecAddScalarVU0;
        vmp_VecSubScalar =      vmp_VecSubScalarVU;
        vmp_VecSubScalar0 =     vmp_VecSubScalarVU0;
      }
    else
      {
        vmp_VecAdd =            vmp_VecAddGeneric;
        vmp_VecAdd0 =           vmp_VecAddGeneric;
        vmp_VecSub =            vmp_VecSubGeneric;
        vmp_VecSub0 =           vmp_VecSubGeneric;
        vmp_VecAddScalar =      vmp_VecAddScalarGeneric;
        vmp_VecAddScalar0 =     vmp_VecAddScalarGeneric;
        vmp_VecSubScalar =      vmp_VecSubScalarGeneric;
        vmp_VecSubScalar0 =     vmp_VecSubScalarGeneric;
      }
}





