
//
//  QVmd3DVU.c - (Quad) 3D Vector Multiplication & Division
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 3 Oct 2001 - James Leiterman
//
//  Last Modification
//
//  stab=4
//


#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries
    
#include    <assert.h>
#include    <stdio.h>

#ifdef CC_VMP_PS2
#ifndef HOMEBREW
#include    <libvu0.h>
#endif
#endif

    //  Book Development Library Modules

#include    "CpuAsm.h"                                  // CPU
#include    "vmp3D.h"
#include    "vmp3DEx.h"
#include    "QVmd3D.h"                                  // 3D Math
#include    "QVmd3DEx.h"



#ifdef HOMEBREW
typedef float       sceVu0FVECTOR[4] __attribute__((aligned (16)));
#define USE_ASM
#endif



//
//  (Quad) Vector Product (Multiplication) - Unaligned
//

void vmp_QVecMulVU0(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB)
{   
    sceVu0FVECTOR vD, vA, vB;
    
    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);
    ASSERT_PTR4(pvB);
    
    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
    vA[3] = pvA->w;
        
    vB[0] = pvB->x;
    vB[1] = pvB->y;
    vB[2] = pvB->z;
    vB[3] = pvB->w;

#ifdef USE_ASM
    asm __volatile__("

    lqc2    vf4,0x0(%1)         # A
    lqc2    vf5,0x0(%2)         # B
    
    vmul.xyzw   vf6,vf4,vf5     # D.xyzw = A.xyzw * B.xyzw

    sqc2    vf6,0x0(%0)

    ":
     : "r" (vD) , "r" (vA), "r" (vB) 
    );

#else
    sceVu0MulVector(vD, vA, vB);
#endif

    pvD->x = vD[0];
    pvD->y = vD[1];
    pvD->z = vD[2];
    pvD->w = vD[3];
}



//
//  (Quad) Vector Product (Multiplication) - Aligned
//

void vmp_QVecMulVU(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB)
{   
    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);
    ASSERT_PTR16(pvB);

#ifdef USE_ASM
    asm __volatile__("

    lqc2    vf4,0x0(%1)         # A
    lqc2    vf5,0x0(%2)         # B
    
    vmul.xyzw   vf6,vf4,vf5     # D.xyzw = A.xyzw * B.xyzw

    sqc2    vf6,0x0(%0)

    ":
     : "r" (pvD) , "r" (pvA), "r" (pvB) 
    );

#else
    sceVu0MulVector(*(sceVu0FVECTOR *)pvD, *(sceVu0FVECTOR *)pvA, *(sceVu0FVECTOR *)pvB);
#endif
}



//______________________________________________________
//
//  (Quad) Vector Product-Sum (Multiplication-Add) - Unaligned
//

void vmp_QVecMAddVU0(vmp3DQVector * const pvD, const vmp3DQVector * const pvA,
              const vmp3DQVector * const pvB, const vmp3DQVector * const pvC)
{   
    sceVu0FVECTOR vD, vA, vB, vC;
    
    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);
    ASSERT_PTR4(pvB);
    ASSERT_PTR4(pvC);
        
    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
    vA[3] = pvA->w;
        
    vB[0] = pvB->x;
    vB[1] = pvB->y;
    vB[2] = pvB->z;
    vB[3] = pvB->w;

    vC[0] = pvC->x;
    vC[1] = pvC->y;
    vC[2] = pvC->z;
    vC[3] = pvC->w;

#ifdef USE_ASM
    asm __volatile__("

    lqc2    vf4,0x0(%1)             # A
    lqc2    vf5,0x0(%2)             # B
    lqc2    vf7,0x0(%3)             # C
        
    vmul.xyzw   vf6,vf4,vf5     # D.xyzw = (A.xyzw * B.xyzw)
    vadd.xyzw   vf6,vf6,vf7     # D.xyzw += C.xyzw

    sqc2    vf6,0x0(%0)

    ":
     : "r" (vD) , "r" (vA), "r" (vB), "r" (vC)
    );

#else
    sceVu0MulVector(vD, vA, vB);
    sceVu0AddVector(vD, vD, vC);
#endif

    pvD->x = vD[0];
    pvD->y = vD[1];
    pvD->z = vD[2];
    pvD->w = vD[3];
}



//
//  (Quad) Vector Product Sum (Multiplication-Add) - Aligned
//

void vmp_QVecMAddVU(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, 
               const vmp3DQVector * const pvB, const vmp3DQVector * const pvC)
{   
    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);
    ASSERT_PTR16(pvB);
    ASSERT_PTR16(pvC);

#ifdef USE_ASM
    asm __volatile__("

    lqc2    vf4,0x0(%1)         # A
    lqc2    vf5,0x0(%2)         # B
    lqc2    vf7,0x0(%3)         # C
        
    vmul.xyzw   vf6,vf4,vf5     # D.xyzw = (A.xyzw * B.xyzw)
    vadd.xyzw   vf6,vf6,vf7     # D.xyzw += C.xyzw

    sqc2    vf6,0x0(%0)

    ":
     : "r" (pvD) , "r" (pvA), "r" (pvB), "r" (pvC) 
    );

#else
    sceVu0MulVector(*(sceVu0FVECTOR *)pvD, *(sceVu0FVECTOR *)pvA, *(sceVu0FVECTOR *)pvB);
    sceVu0AddVector(*(sceVu0FVECTOR *)pvD, *(sceVu0FVECTOR *)pvD, *(sceVu0FVECTOR *)pvC);
#endif
}



//______________________________________________________
//
//  (Quad) Vector Product-Dif (Multiplication-Sub) - Unaligned
//

void vmp_QVecMSubVU0(vmp3DQVector * const pvD, const vmp3DQVector * const pvA,
              const vmp3DQVector * const pvB, const vmp3DQVector * const pvC)
{   
    sceVu0FVECTOR vD, vA, vB, vC;
    
    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);
    ASSERT_PTR4(pvB);
    ASSERT_PTR4(pvC);
        
    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
    vA[3] = pvA->w;
        
    vB[0] = pvB->x;
    vB[1] = pvB->y;
    vB[2] = pvB->z;
    vB[3] = pvB->w;

    vC[0] = pvC->x;
    vC[1] = pvC->y;
    vC[2] = pvC->z;
    vC[3] = pvC->w;

#ifdef USE_ASM
    asm __volatile__("

    lqc2    vf4,0x0(%1)         # A
    lqc2    vf5,0x0(%2)         # B
    lqc2    vf7,0x0(%3)         # C
        
    vmul.xyzw   vf6,vf4,vf5     # D.xyzw = (A.xyzw * B.xyzw)
    vsub.xyzw   vf6,vf6,vf7     # D.xyzw -= C.xyzw

    sqc2    vf6,0x0(%0)

    ":
     : "r" (vD) , "r" (vA), "r" (vB), "r" (vC)
    );

#else
    sceVu0MulVector(vD, vA, vB);
    sceVu0SubVector(vD, vD, vC);
#endif

    pvD->x = vD[0];
    pvD->y = vD[1];
    pvD->z = vD[2];
    pvD->w = vD[3];
}



//
//  (Quad) Vector Product Dif (Multiplication-Sub) - Aligned
//

void vmp_QVecMSubVU(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, 
               const vmp3DQVector * const pvB, const vmp3DQVector * const pvC)
{   
    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);
    ASSERT_PTR16(pvB);
    ASSERT_PTR16(pvC);

#ifdef USE_ASM
    asm __volatile__("

    lqc2    vf4,0x0(%1)         #A
    lqc2    vf5,0x0(%2)         #B
    lqc2    vf7,0x0(%3)         #C
        
    vmul.xyzw   vf6,vf4,vf5     # D.xyzw = (A.xyzw * B.xyzw)
    vsub.xyzw   vf6,vf6,vf7     # D.xyzw -= C.xyzw

    sqc2    vf6,0x0(%0)

    ":
     : "r" (pvD) , "r" (pvA), "r" (pvB), "r" (pvC) 
    );

#else
    sceVu0MulVector(*(sceVu0FVECTOR *)pvD, *(sceVu0FVECTOR *)pvA, *(sceVu0FVECTOR *)pvB);
    sceVu0SubVector(*(sceVu0FVECTOR *)pvD, *(sceVu0FVECTOR *)pvD, *(sceVu0FVECTOR *)pvC);
#endif
}



//______________________________________________________
//
//  (Quad) Vector (Division) - Unaligned
//

void vmp_QVecDivVU0(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, float fScalar)
{   
    sceVu0FVECTOR vD, vA;
    
    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);
    ASSERT_ZERO(fScalar);

    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
    vA[3] = pvA->w;

#ifdef USE_ASM
    asm __volatile__("

    lqc2    vf4,0x0(%1)         #A
    mfc1    $8,%2               #Scalar
    qmtc2    $8,vf5             #vf5=  xfer FP from EE to VU
    
    vdiv    Q, vf0w, vf5x       # Q = 1 / scalar
    vwaitq
    
    vmulq.xyzw  vf6,vf4,Q       # D.xyzw = A.xyzw / scalar
    
    sqc2    vf6,0x0(%0)

    ":
     : "r" (vD) , "r" (vA), "f" (fScalar) 
     : "$8"
    );

#else
    sceVu0DivVector(vD, vA, fScalar);
#endif

    pvD->x = vD[0];
    pvD->y = vD[1];
    pvD->z = vD[2];
    pvD->w = vD[3];
}



//
//  (Quad) Vector Product (Multiplication) - Aligned
//

void vmp_QVecDivVU(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, float fScalar)
{   
    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);
    ASSERT_ZERO(fScalar);

#ifdef USE_ASM
    asm __volatile__("

    lqc2    vf4,0x0(%1)         #A
    mfc1    $8,%2               #Scalar
    qmtc2    $8,vf5             #vf5=  xfer FP from EE to VU
    
    vdiv    Q, vf0w, vf5x       # Q = 1 / scalar
    vwaitq
    
    vmulq.xyzw  vf6,vf4,Q       # D.xyzw = A.xyzw / scalar
    
    sqc2    vf6,0x0(%0)

    ":
     : "r" (pvD) , "r" (pvA), "f" (fScalar) 
     : "$8"
    );

#else
    sceVu0DivVector(*(sceVu0FVECTOR *)pvD, *(sceVu0FVECTOR *)pvA, fScalar);
#endif
}



//______________________________________________________
//
//
//  (Quad) Scale Vector w/Scalar - Unaligned
//

void vmp_QVecScaleVU0(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, float fScalar )
{
    sceVu0FVECTOR vD, vA;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);

    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
    vA[3] = pvA->w;
        
#ifdef USE_ASM
    asm __volatile__("

    lqc2    vf4,0x0(%1)         #A
    mfc1    $8,%2               #Scalar
    qmtc2    $8,vf5             #vf5=  xfer FP from EE to VU
    
    vmulx.xyzw  vf6,vf4,vf5x    # D.xyzw = A.xyzw * Scalar

    sqc2    vf6,0x0(%0)

    ":
     : "r" (vD) , "r" (vA), "f" (fScalar) 
    );

#else
    sceVu0ScaleVector(vD, vA, fScalar);
#endif

    pvD->x = vD[0];
    pvD->y = vD[1];
    pvD->z = vD[2];
    pvD->w = vD[3];
}



//
//  (Quad) Scale Vector w/Scalar - Aligned
//

void vmp_QVecScaleVU(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, float fScalar )
{
    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);

#ifdef USE_ASM
    asm __volatile__("

    lqc2    vf4,0x0(%1)         #A
    mfc1    $8,%2               #Scalar
    qmtc2    $8,vf5             #vf5=  xfer FP from EE to VU
    
    vmulx.xyzw  vf6,vf4,vf5x    # D.xyzw = A.xyzw * Scalar

    sqc2    vf6,0x0(%0)

    ":
     : "r" (pvD) , "r" (pvA), "f" (fScalar) 
    );

#else
    sceVu0ScaleVector(*(sceVu0FVECTOR *)pvD, *(sceVu0FVECTOR *)pvA, fScalar);
#endif
}



//______________________________________________________
//
//  Dot Product - UnAligned
//
// f = pvA->x * pvB->x + pvA->y * pvB->y + pvA->z * pvB->z;
//
// NOTE: Not very efficient, should do four dot products in parallel!

void vmp_QDotProductVU0(float * const pfD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB)
{
#ifdef USE_ASM
    sceVu0FVECTOR vD, vA, vB;

    ASSERT_PTR4(pfD);
    ASSERT_PTR4(pvA);
    ASSERT_PTR4(pvB);

    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
        
    vB[0] = pvB->x;
    vB[1] = pvB->y;
    vB[2] = pvB->z;

    asm __volatile__("

    lqc2    vf4,0x0(%1)         # A
    lqc2    vf5,0x0(%2)         # B
    
    vmul.xyz    vf6,vf4,vf5     # D.x = A.xyz * B.xyz
    vaddy.x     vf6,vf6,vf6     # D.x += D.y
    vaddz.x     vf6,vf6,vf6     # D.x += D.z
    
    sqc2    vf6,0x0(%0)

    ":
     : "r" (vD) , "r" (vA), "r" (vB) 
    );

    *pfD = vD[0];
    
#else
    sceVu0FVECTOR vA, vB;

    ASSERT_PTR4(pfD);
    ASSERT_PTR4(pvA);
    ASSERT_PTR4(pvB);

    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
    vA[3] = 0.0f;
        
    vB[0] = pvB->x;
    vB[1] = pvB->y;
    vB[2] = pvB->z;
    vB[3] = 0.0f;

    *pfD = sceVu0InnerProduct(vA, vB);
#endif
}



//  Dot Product - (Un)Aligned
//
// f = pvA->x * pvB->x + pvA->y * pvB->y + pvA->z * pvB->z;
//
// NOTE: Not very efficient, should do four dot products in parallel!

void vmp_QDotProductVU(float * const pfD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB)
{
#ifdef USE_ASM
    sceVu0FVECTOR vD;

    ASSERT_PTR4(pfD);
    ASSERT_PTR16(pvA);
    ASSERT_PTR16(pvB);

    asm __volatile__("

    lqc2    vf4,0x0(%1)         # A
    lqc2    vf5,0x0(%2)         # B
    
    vmul.xyz    vf6,vf4,vf5     # D.x = A.xyz * B.xyz
    vaddy.x     vf6,vf6,vf6     # D.x += D.y
    vaddz.x     vf6,vf6,vf6     # D.x += D.z
    
    sqc2    vf6,0x0(%0)

    ":
     : "r" (vD) , "r" (pvA), "r" (pvB) 
    );

    *pfD = vD[0];
    
#else
    ASSERT_PTR4(pfD);
    ASSERT_PTR16(pvA);
    ASSERT_PTR16(pvB);

    *pfD = sceVu0InnerProduct( *(sceVu0FVECTOR *)pvA, *(sceVu0FVECTOR *)pvB);
#endif
}



//______________________________________________________
//
//  (Quad Vector) Cross Product - Unaligned
//
//  x = pvA->y * pvB->z - pvA->z * pvB->y;    1 2   2 1
//  y = pvA->z * pvB->x - pvA->x * pvB->z;    2 0   0 2
//  z = pvA->x * pvB->y - pvA->y * pvB->x;    0 1   1 0
//  w = 0.0

void vmp_QCrossProductVU0(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB)
{   
    sceVu0FVECTOR vD, vA, vB;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);
    ASSERT_PTR4(pvB);

    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
    vA[3] = 0.0f;
        
    vB[0] = pvB->x;
    vB[1] = pvB->y;
    vB[2] = pvB->z;
    vB[3] = 0.0f;

#ifdef USE_ASM
    asm __volatile__("

    lqc2    vf4,0x0(%1)         # A
    lqc2    vf5,0x0(%2)         # B

    vopmula.xyz ACC,vf4,vf5
    vopmsub.xyz vf6,vf5,vf4     #
    vsub.w  vf6,vf6,vf6         # D.w = 0.0
    
    sqc2    vf6,0x0(%0)

    ":
     : "r" (vD) , "r" (vA), "r" (vB) 
    );

#else
    sceVu0OuterProduct(vD, vA, vB);
#endif
    
    pvD->x = vD[0];
    pvD->y = vD[1];
    pvD->z = vD[2];
    pvD->w = vD[3]; // = 0.0
}



//  Cross Product - Aligned
//
//  x = pvA->y * pvB->z - pvA->z * pvB->y;    1 2   2 1
//  y = pvA->z * pvB->x - pvA->x * pvB->z;    2 0   0 2
//  z = pvA->x * pvB->y - pvA->y * pvB->x;    0 1   1 0
//  w = 0.0

void vmp_QCrossProductVU(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB)
{   
    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);
    ASSERT_PTR16(pvB);

#ifdef USE_ASM
    asm __volatile__("

    lqc2    vf4,0x0(%1)         # A
    lqc2    vf5,0x0(%2)         # B

    vopmula.xyz ACC,vf4,vf5
    vopmsub.xyz vf6,vf5,vf4     #
    vsub.w  vf6,vf6,vf6         # D.w = 0.0
    
    sqc2    vf6,0x0(%0)

    ":
     : "r" (pvD) , "r" (pvA), "r" (pvB) 
    );

#else
    sceVu0OuterProduct( *(sceVu0FVECTOR *)pvD, *(sceVu0FVECTOR *)pvA, *(sceVu0FVECTOR *)pvB);
#endif
}




//______________________________________________________
//
//  Setup Vec3D module
//
//  NOTE: This function sets up the function vectors so the best function
//      is attached to the vector for the processor running.
//

void QVmd3DSetup( const CpuInfo * const pcinfo )
{
    ASSERT_PTR4(pcinfo);
    
    if ( CPUBITS_VU0 & pcinfo->nBits )
      {
        vmp_QVecMul =           vmp_QVecMulVU;
        vmp_QVecMul0 =          vmp_QVecMulVU0;
        vmp_QVecMAdd =          vmp_QVecMAddVU;
        vmp_QVecMAdd0 =         vmp_QVecMAddVU0;
        vmp_QVecMSub =          vmp_QVecMSubVU;
        vmp_QVecMSub0 =         vmp_QVecMSubVU0;
        vmp_QVecScale =         vmp_QVecScaleVU;
        vmp_QVecScale0 =        vmp_QVecScaleVU0;
        vmp_QVecDiv =           vmp_QVecDivVU;
        vmp_QVecDiv0 =          vmp_QVecDivVU0;
        vmp_QVecDivFast =       vmp_QVecDivVU;      // Note: No Fast division
        vmp_QVecDivFast0 =      vmp_QVecDivVU0;     //      so use standard!
        vmp_QDotProduct =       vmp_QDotProductVU;
        vmp_QDotProduct0 =      vmp_QDotProductVU0;
        vmp_QCrossProduct =     vmp_QCrossProductVU;
        vmp_QCrossProduct0 =    vmp_QCrossProductVU0;
      }

    else
      {
        vmp_QVecMul =           vmp_QVecMulGeneric;
        vmp_QVecMul0 =          vmp_QVecMulGeneric;
        vmp_QVecMAdd =          vmp_QVecMAddGeneric;
        vmp_QVecMAdd0 =         vmp_QVecMAddGeneric;
        vmp_QVecMSub =          vmp_QVecMSubGeneric;
        vmp_QVecMSub0 =         vmp_QVecMSubGeneric;
        vmp_QVecScale =         vmp_QVecScaleGeneric;
        vmp_QVecScale0 =        vmp_QVecScaleGeneric;
        vmp_QVecDiv =           vmp_QVecDivGeneric;
        vmp_QVecDiv0 =          vmp_QVecDivGeneric;
        vmp_QVecDivFast =       vmp_QVecDivGeneric;
        vmp_QVecDivFast0 =      vmp_QVecDivGeneric;
        vmp_QDotProduct =       vmp_QDotProductGeneric;
        vmp_QDotProduct0 =      vmp_QDotProductGeneric;
        vmp_QCrossProduct =     vmp_QCrossProductGeneric;
        vmp_QCrossProduct0 =    vmp_QCrossProductGeneric;
      }
}



