
//
//  Vmd3DVU.c - 3D Vector Multiplication & Division
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  NOTE: Because these a 3 float Vectors the 4th float has to be handled as 
//      a special case thus slowing the code down. This is the same as an AoS 
//      orientation. Instead a SoA should be used in conjunction with the quad
//      set of functions.
//
//  Created - 3 Oct 2001 - James Leiterman
//
//  Last Modification
//
//  stab=4
//


#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries
    
#include    <assert.h>
#include    <stdio.h>

#ifdef CC_VMP_PS2
#ifndef HOMEBREW
#include    <libvu0.h>
#endif
#endif

    //  Book Development Library Modules

#include    "CpuAsm.h"                                  // CPU
#include    "vmp3D.h"
#include    "vmp3DEx.h"
#include    "Vmd3D.h"                                   // 3D Math
#include    "Vmd3DEx.h"



#ifdef HOMEBREW
typedef float       sceVu0FVECTOR[4] __attribute__((aligned (16)));
#define USE_ASM
#endif



//
//  Vector Product (Multiplication) - UnAligned
//

void vmp_VecMulVU0(vmp3DVector * const pvD, const vmp3DVector * const pvA, const vmp3DVector * const pvB)
{   
    sceVu0FVECTOR vD, vA, vB;
    
    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);
    ASSERT_PTR4(pvB);
    
    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
        
    vB[0] = pvB->x;
    vB[1] = pvB->y;
    vB[2] = pvB->z;
    
#ifdef USE_ASM
    asm __volatile__("

    lqc2    vf4,0x0(%1)         # A
    lqc2    vf5,0x0(%2)         # B
    
    vmul.xyz    vf6,vf4,vf5     # D.xyz = A.xyz * B.xyz

    sqc2    vf6,0x0(%0)

    ":
     : "r" (vD) , "r" (vA), "r" (vB) 
    );

#else
    sceVu0MulVector( vD, vA, vB );
#endif

    pvD->x = vD[0];
    pvD->y = vD[1];
    pvD->z = vD[2];
}



//
//  Vector Product (Multiplication) - Aligned
//

void vmp_VecMulVU(vmp3DVector * const pvD, const vmp3DVector * const pvA, const vmp3DVector * const pvB)
{   
#ifdef USE_ASM
    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);
    ASSERT_PTR16(pvB);
    
    asm __volatile__("

    lqc2    vf4,0x0(%1)         # A
    lqc2    vf5,0x0(%2)         # B
    lqc2    vf6,0x0(%0)         # D
    
    vmul.xyz    vf6,vf4,vf5     # D.xyz = A.xyz * B.xyz

    sqc2    vf6,0x0(%0)

    ":
     : "r" (pvD) , "r" (pvA), "r" (pvB) 
    );

#else
    sceVu0FVECTOR vD;
    
    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);
    ASSERT_PTR16(pvB);
    
    sceVu0MulVector(vD, *(sceVu0FVECTOR *)pvA, *(sceVu0FVECTOR *)pvB);

    sceVu0CopyVectorXYZ(*(sceVu0FVECTOR *)pvD, vD );
#endif
}



//______________________________________________________
//
//  Vector Product Sum (Multiplication-Add) - UnAligned
//

void vmp_VecMAddVU0(vmp3DVector * const pvD, const vmp3DVector * const pvA, 
             const vmp3DVector * const pvB, const vmp3DVector * const pvC)
{   
    sceVu0FVECTOR vD, vA, vB, vC;
    
    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);
    ASSERT_PTR4(pvB);
    ASSERT_PTR4(pvC);
        
    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
        
    vB[0] = pvB->x;
    vB[1] = pvB->y;
    vB[2] = pvB->z;
    
    vC[0] = pvC->x;
    vC[1] = pvC->y;
    vC[2] = pvC->z;
    
#ifdef USE_ASM
    asm __volatile__("

    lqc2    vf4,0x0(%1)             # A
    lqc2    vf5,0x0(%2)             # B
    lqc2    vf7,0x0(%3)             # C
        
    vmul.xyz    vf6,vf4,vf5     # D.xyz = (A.xyz * B.xyz)
    vadd.xyz    vf6,vf6,vf7     # D.xyz += C.xyz

    sqc2    vf6,0x0(%0)

    ":
     : "r" (vD) , "r" (vA), "r" (vB), "r" (vC)
    );

#else
    sceVu0MulVector( vD, vA, vB );
    sceVu0AddVector( vD, vD, vC );
#endif

    pvD->x = vD[0];
    pvD->y = vD[1];
    pvD->z = vD[2];
}



//
//  Vector Product Sum (Multiplication-Add) - Aligned
//

void vmp_VecMAddVU(vmp3DVector * const pvD, const vmp3DVector * const pvA,
              const vmp3DVector * const pvB, const vmp3DVector * const pvC)
{   
#ifdef USE_ASM
    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);
    ASSERT_PTR16(pvB);
    ASSERT_PTR16(pvC);
    
    asm __volatile__("

    lqc2    vf4,0x0(%1)         # A
    lqc2    vf5,0x0(%2)         # B
    lqc2    vf7,0x0(%3)         # C
    lqc2    vf6,0x0(%0)         # D
            
    vmul.xyz    vf6,vf4,vf5     # D.xyz = (A.xyz * B.xyz)
    vadd.xyz    vf6,vf6,vf7     # D.xyz += C.xyz

    sqc2    vf6,0x0(%0)

    ":
     : "r" (pvD) , "r" (pvA), "r" (pvB), "r" (pvC) 
    );

#else
    sceVu0FVECTOR vD;
    
    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);
    ASSERT_PTR16(pvB);
    ASSERT_PTR16(pvC);
    
    sceVu0MulVector(vD, *(sceVu0FVECTOR *)pvA, *(sceVu0FVECTOR *)pvB);
    sceVu0AddVector(vD, vD, *(sceVu0FVECTOR *)pvC);

    sceVu0CopyVectorXYZ(*(sceVu0FVECTOR *)pvD, vD );
#endif
}



//______________________________________________________
//
//  Vector Product Dif (Multiplication-Sub) - UnAligned
//

void vmp_VecMSubVU0(vmp3DVector * const pvD, const vmp3DVector * const pvA, 
             const vmp3DVector * const pvB, const vmp3DVector * const pvC)
{   
    sceVu0FVECTOR vD, vA, vB, vC;
    
    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);
    ASSERT_PTR4(pvB);
    ASSERT_PTR4(pvC);
        
    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
    vA[3] = 0.0f;
        
    vB[0] = pvB->x;
    vB[1] = pvB->y;
    vB[2] = pvB->z;
    vB[3] = 0.0f;
    
    vC[0] = pvC->x;
    vC[1] = pvC->y;
    vC[2] = pvC->z;
    vC[3] = 0.0f;
    
    sceVu0MulVector( vD, vA, vB );
    sceVu0SubVector( vD, vD, vC );

    pvD->x = vD[0];
    pvD->y = vD[1];
    pvD->z = vD[2];
}



//
//  Vector Product Dif (Multiplication-Sub) - Aligned
//

void vmp_VecMSubVU(vmp3DVector * const pvD, const vmp3DVector * const pvA,
              const vmp3DVector * const pvB, const vmp3DVector * const pvC)
{   
    sceVu0FVECTOR vD;
    
    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);
    ASSERT_PTR16(pvB);
    ASSERT_PTR16(pvC);
        
    sceVu0MulVector(vD, *(sceVu0FVECTOR *)pvA, *(sceVu0FVECTOR *)pvB);
    sceVu0SubVector(vD, vD, *(sceVu0FVECTOR *)pvC);

    sceVu0CopyVectorXYZ(*(sceVu0FVECTOR *)pvD, vD );
}



//______________________________________________________
//
//  Scale Vector w/Scalar - Unaligned
//

void vmp_VecScaleVU0(vmp3DVector * const pvD, const vmp3DVector * const pvA, float fScalar )
{
    sceVu0FVECTOR vD, vA;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);
    
    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
        
    sceVu0ScaleVectorXYZ(vD, vA, fScalar);

    pvD->x = vD[0];
    pvD->y = vD[1];
    pvD->z = vD[2];
}




//
//  Scale Vector w/Scalar - aligned
//

void vmp_VecScaleVU(vmp3DVector * const pvD, const vmp3DVector * const pvA, float fScalar )
{
    sceVu0FVECTOR vD;

    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);
    
    sceVu0ScaleVectorXYZ(vD, *(sceVu0FVECTOR *)pvA, fScalar);
        // Note: .W gets altered so it's double handled here!

    sceVu0CopyVectorXYZ(*(sceVu0FVECTOR *)pvD, vD );
}



//______________________________________________________
//
//  Vector Scalar Division - Unaligned
//

void vmp_VecDivVU0(vmp3DVector * const pvD, const vmp3DVector * const pvA, float fScalar )
{
    sceVu0FVECTOR vD, vA;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);
    ASSERT_ZERO(fScalar);
        
    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
        
    sceVu0DivVectorXYZ(vD, vA, fScalar);

    pvD->x = vD[0];
    pvD->y = vD[1];
    pvD->z = vD[2];
}




//
//  Vector Scalar Division - aligned
//

void vmp_VecDivVU(vmp3DVector * const pvD, const vmp3DVector * const pvA, float fScalar )
{
    sceVu0FVECTOR vD;

    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);
    ASSERT_ZERO(fScalar);

    sceVu0DivVectorXYZ(vD, *(sceVu0FVECTOR *)pvA, fScalar);
        // Note: .W gets altered so it's double handled here!

    sceVu0CopyVectorXYZ(*(sceVu0FVECTOR *)pvD, vD );
}



//______________________________________________________
//
//  Dot Product - UnAligned
//
// f = pvA->x * pvB->x + pvA->y * pvB->y + pvA->z * pvB->z;
//
// NOTE: Not very efficient, should do four dot products in parallel!

void vmp_DotProductVU0(float * const pfD, const vmp3DVector * const pvA, const vmp3DVector * const pvB)
{
    sceVu0FVECTOR vA, vB;

    ASSERT_PTR4(pfD);
    ASSERT_PTR4(pvA);
    ASSERT_PTR4(pvB);

    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
    vA[3] = 0.0f;
        
    vB[0] = pvB->x;
    vB[1] = pvB->y;
    vB[2] = pvB->z;
    vB[3] = 0.0f;

    *pfD = sceVu0InnerProduct(vA, vB);
}



//  Dot Product - (Un)Aligned
//
// f = pvA->x * pvB->x + pvA->y * pvB->y + pvA->z * pvB->z;
//
// NOTE: Not very efficient, should do four dot products in parallel!

void vmp_DotProductVU(float * const pfD, const vmp3DVector * const pvA, const vmp3DVector * const pvB)
{
    ASSERT_PTR4(pfD);
    ASSERT_PTR16(pvA);
    ASSERT_PTR16(pvB);

    *pfD = sceVu0InnerProduct( *(sceVu0FVECTOR *)pvA, *(sceVu0FVECTOR *)pvB);
}



//______________________________________________________
//
//  Cross Product - Unaligned
//
//  x = pvA->y * pvB->z - pvA->z * pvB->y;    1 2   2 1
//  y = pvA->z * pvB->x - pvA->x * pvB->z;    2 0   0 2
//  z = pvA->x * pvB->y - pvA->y * pvB->x;    0 1   1 0

void vmp_CrossProductVU0(vmp3DVector * const pvD, const vmp3DVector * const pvA, const vmp3DVector * const pvB)
{   
    sceVu0FVECTOR vD, vA, vB;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);
    ASSERT_PTR4(pvB);

    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
    vA[3] = 0.0f;
        
    vB[0] = pvB->x;
    vB[1] = pvB->y;
    vB[2] = pvB->z;
    vB[3] = 0.0f;

    sceVu0OuterProduct(vD, vA, vB);
    
    pvD->x = vD[0];
    pvD->y = vD[1];
    pvD->z = vD[2];
}



//  Cross Product - Aligned
//
//  x = pvA->y * pvB->z - pvA->z * pvB->y;    1 2   2 1
//  y = pvA->z * pvB->x - pvA->x * pvB->z;    2 0   0 2
//  z = pvA->x * pvB->y - pvA->y * pvB->x;    0 1   1 0

void vmp_CrossProductVU(vmp3DVector * const pvD, const vmp3DVector * const pvA, const vmp3DVector * const pvB)
{   
    sceVu0FVECTOR vD;

    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);
    ASSERT_PTR16(pvB);

    sceVu0OuterProduct( vD, *(sceVu0FVECTOR *)pvA, *(sceVu0FVECTOR *)pvB);
        // Note: OuterProduct sets .W to 0.0
    sceVu0CopyVectorXYZ(*(sceVu0FVECTOR *)pvD, vD );
}



//______________________________________________________
//
//  Setup Vec3D module
//
//  NOTE: This function sets up the function vectors so the best function
//      is attached to the vector for the processor running.
//

void Vmd3DSetup( const CpuInfo * const pcinfo )
{
    ASSERT_PTR4(pcinfo);
    
    if ( CPUBITS_VU0 & pcinfo->nBits )
      {
        vmp_VecMul =            vmp_VecMulVU;
        vmp_VecMul0 =           vmp_VecMulVU0;
        vmp_VecMAdd =           vmp_VecMAddVU;
        vmp_VecMAdd0 =          vmp_VecMAddVU0;
        vmp_VecMSub =           vmp_VecMSubVU;
        vmp_VecMSub0 =          vmp_VecMSubVU0;
        vmp_VecDiv =            vmp_VecDivVU;
        vmp_VecDiv0 =           vmp_VecDivVU0;
        vmp_VecDivFast =        vmp_VecDivVU;
        vmp_VecDivFast0 =       vmp_VecDivVU0;
        vmp_VecScale =          vmp_VecScaleVU;
        vmp_VecScale0 =         vmp_VecScaleVU0;
        vmp_DotProduct =        vmp_DotProductVU;
        vmp_DotProduct0 =       vmp_DotProductVU0;
        vmp_CrossProduct =      vmp_CrossProductVU;
        vmp_CrossProduct0 =     vmp_CrossProductVU0;
      }

    else
      {
        vmp_VecMul =            vmp_VecMulGeneric;
        vmp_VecMul0 =           vmp_VecMulGeneric;
        vmp_VecMAdd =           vmp_VecMAddGeneric;
        vmp_VecMAdd0 =          vmp_VecMAddGeneric;
        vmp_VecMSub =           vmp_VecMSubGeneric;
        vmp_VecMSub0 =          vmp_VecMSubGeneric;
        vmp_VecDiv =            vmp_VecDivGeneric;
        vmp_VecDiv0 =           vmp_VecDivGeneric;
        vmp_VecDivFast =        vmp_VecDivGeneric;
        vmp_VecDivFast0 =       vmp_VecDivGeneric;
        vmp_VecScale =          vmp_VecScaleGeneric;
        vmp_VecScale0 =         vmp_VecScaleGeneric;
        vmp_DotProduct =        vmp_DotProductGeneric;
        vmp_DotProduct0 =       vmp_DotProductGeneric;
        vmp_CrossProduct =      vmp_CrossProductGeneric;
        vmp_CrossProduct0 =     vmp_CrossProductGeneric;
      }
}



