
//
//  QVsf3DVU.c - 3D (Quad) Vector Special Function
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 3 Oct 2001 - James Leiterman
//
//  Last Modification
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries
    
#include    <assert.h>
#include    <stdio.h>
#include    <math.h>

#ifdef CC_VMP_PS2
#include    <libvu0.h>
#endif

    //  Book Development Library Modules

#include    "CpuAsm.h"                                  // CPU
#include    "vmp3D.h"
#include    "vmp3DEx.h"

#include    "QVsf3D.h"                                  // 3D Math
#include    "QVsf3DEx.h"


#ifdef HOMEBREW
typedef float       sceVu0FVECTOR[4] __attribute__((aligned (16)));
#define USE_ASM
#endif



//
//  (Quad) Vector Square Root - Unaligned
//
//  pvD->x = sqrtf( pvA->x );
//  pvD->y = sqrtf( pvA->y );
//  pvD->z = sqrtf( pvA->z );
//  pvD->w = sqrtf( pvA->w );
    
void vmp_QVecSqrtVU0(vmp3DQVector * const pvD, const vmp3DQVector * const pvA)
{
    sceVu0FVECTOR vA, vD;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);

    ASSERT_FNEG(pvA->x);
    ASSERT_FNEG(pvA->y);
    ASSERT_FNEG(pvA->z);
    ASSERT_FNEG(pvA->w);

    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
    vA[3] = pvA->w;
    
    asm __volatile__("

    lqc2    vf5,0x0(%1)         #pvA

    vsqrt Q,vf5x                # q=sqrt(Ax)
    vwaitq      
    vaddq.x vf4x, vf0x, Q       #  =0+q

    vsqrt Q,vf5y                # q=sqrt(Ay)
    vwaitq      
    vaddq.y vf4y, vf0y, Q       #  =0+q
     
    vsqrt Q,vf5z                # q=sqrt(Az)
    vwaitq      
    vaddq.z vf4z, vf0z, Q       #  =0+q

    vsqrt Q,vf5w                # q=sqrt(Aw)
    vwaitq      
    vaddq.w vf4w, vf0w, Q       #  =1+q

    vsub.w vf4w, vf4w, vf0w     # -1
    
    sqc2    vf4,0x0(%0)         # write Vector XYZW
  
    ": : "r" (vD) , "r" (vA));

    pvD->x = vD[0];
    pvD->y = vD[1];
    pvD->z = vD[2];
    pvD->w = vD[3];
}


//
//  (Quad) Vector Square Root - Aligned
//
    
void vmp_QVecSqrtVU(vmp3DQVector * const pvD, const vmp3DQVector * const pvA)
{
    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);

    ASSERT_FNEG(pvA->x);
    ASSERT_FNEG(pvA->y);
    ASSERT_FNEG(pvA->z);
    ASSERT_FNEG(pvA->w);

    asm __volatile__("
    lqc2    vf5,0x0(%1)         #pvA

    vsqrt Q,vf5x                # q=sqrt(Ax)
    vwaitq      
    vaddq.x vf4x, vf0x, Q       #  =0+q

    vsqrt Q,vf5y                # q=sqrt(Ay)
    vwaitq      
    vaddq.y vf4y, vf0y, Q       #  =0+q
     
    vsqrt Q,vf5z                # q=sqrt(Az)
    vwaitq      
    vaddq.z vf4z, vf0z, Q       #  =0+q

    vsqrt Q,vf5w                # q=sqrt(Aw)
    vwaitq      
    vaddq.w vf4w, vf0w, Q       #  =1+q

    vsub.w vf4w, vf4w, vf0w     # -1

    sqc2    vf4,0x0(%0)         # write Vector XYZW
  
    ": : "r" (pvD) , "r" (pvA));
}



//______________________________________________________
//
//  (Quad) Vector Minimum - Unaligned
//
//  pvD->x = (pvA->x < pvB->x) ? pvA->x : pvB->x;
//  pvD->y = (pvA->y < pvB->y) ? pvA->y : pvB->y;
//  pvD->z = (pvA->z < pvB->z) ? pvA->z : pvB->z;
//  pvD->w = (pvA->w < pvB->w) ? pvA->w : pvB->w;
    
void vmp_QVecMinVU0( vmp3DQVector * const pvD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB )
{
    sceVu0FVECTOR vA, vB, vD;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);
    ASSERT_PTR4(pvB);
 
    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
    vA[3] = pvA->w;
    
    vB[0] = pvB->x;
    vB[1] = pvB->y;
    vB[2] = pvB->z;
    vB[3] = pvB->w;
    
//#ifdef USE_ASM
    asm __volatile__("
    lqc2    vf6,0x0(%1)         #vA
    lqc2    vf5,0x0(%2)         #vB
    
    vmini.xyzw vf4,vf6,vf5      # min XYZW

    sqc2    vf4,0x0(%0)         # write TMP Vector XYZW
  
    ": : "r" (vD) , "r" (vA), "r" (vB));
//#else
//#endif

    pvD->x = vD[0];
    pvD->y = vD[1];
    pvD->z = vD[2];
    pvD->w = vD[3];
}


//
//  (Quad) Vector Minimum - Aligned
//
    
void vmp_QVecMinVU( vmp3DQVector * const pvD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB )
{
    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);
    ASSERT_PTR16(pvB);
 
//#ifdef USE_ASM
    asm __volatile__("
    lqc2    vf6,0x0(%1)         #pvA
    lqc2    vf5,0x0(%2)         #pvB
    
    vmini.xyzw vf4,vf6,vf5      # min XYZW

    sqc2    vf4,0x0(%0)         #pvD write Vector XYZW
  
    ": : "r" (pvD) , "r" (pvA), "r" (pvB));
//#else
//#endif
}



//______________________________________________________
//
//  (Quad) Vector Maximum - Unaligned
//
//  pvD->x = (pvA->x > pvB->x) ? pvA->x : pvB->x;
//  pvD->y = (pvA->y > pvB->y) ? pvA->y : pvB->y;
//  pvD->z = (pvA->z > pvB->z) ? pvA->z : pvB->z;
//  pvD->w = (pvA->w > pvB->w) ? pvA->w : pvB->w;
    
void vmp_QVecMaxVU0( vmp3DQVector * const pvD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB )
{
    sceVu0FVECTOR vA, vB, vD;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);
    ASSERT_PTR4(pvB);
 
    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
    vA[3] = pvA->w;
    
    vB[0] = pvB->x;
    vB[1] = pvB->y;
    vB[2] = pvB->z;
    vB[3] = pvB->w;
    
    asm __volatile__("
    lqc2    vf6,0x0(%1)         #vA
    lqc2    vf5,0x0(%2)         #vB
    
    vmax.xyzw vf4,vf6,vf5       # max XYZW

    sqc2    vf4,0x0(%0)         # write TMP Vector XYZW
  
    ": : "r" (vD) , "r" (vA), "r" (vB));

    pvD->x = vD[0];
    pvD->y = vD[1];
    pvD->z = vD[2];
    pvD->w = vD[3];
}



//
//  (Quad) Vector Maximum - Aligned
//
    
void vmp_QVecMaxVU( vmp3DQVector * const pvD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB )
{
    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);
    ASSERT_PTR16(pvB);
 
    asm __volatile__("
    lqc2    vf6,0x0(%1)         #pvA
    lqc2    vf5,0x0(%2)         #pvB
    
    vmax.xyzw vf4,vf6,vf5       # max XYZW

    sqc2    vf4,0x0(%0)         #pvD write Vector XYZW
  
    ": : "r" (pvD) , "r" (pvA), "r" (pvB));
}



//______________________________________________________
//
//  (Quad) Vector Magnitude - Unaligned
//
//  *pfD = sqrtf( pvA->x * pvA->x + pvA->y * pvA->y + pvA->z * pvA->z );

void vmp_QVecMagnitudeVU0(float * const pfD, const vmp3DQVector * const pvA)
{
#ifdef USE_ASM
    sceVu0FVECTOR vD, vA;

    ASSERT_PTR4(pfD);
    ASSERT_PTR4(pvA);

    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
    vA[3] = 0.0f;

    asm __volatile__("

    lqc2    vf4,0x0(%1)         # A
        # Dot Product    
    vmul.xyz    vf5,vf4,vf4     # D.x = A.xyz * A.xyz
    vaddy.x     vf5,vf5,vf5     # D.x += D.y
    vaddz.x     vf5,vf5,vf5     # D.x += D.z
        # sqrt
    vsqrt Q,vf5x                # q=sqrt(Dx)
    vwaitq      
    vaddq.x vf4x, vf0x, Q       #  =0+q

    sqc2    vf4,0x0(%0)         # write float

    ":
     : "r" (vD) , "r" (vA) 
    );

    *pfD = vD[0];
    
#else
    sceVu0FVECTOR vA;

    ASSERT_PTR(pfD);
    ASSERT_PTR(pvA);

    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
    vA[3] = 0.0f;
        
    *pfD = sqrt( sceVu0InnerProduct(vA, vA) );
#endif
}



//  (Quad) Vector Magnitude - Aligned
//
//  *pfD = sqrtf( pvA->x * pvA->x + pvA->y * pvA->y + pvA->z * pvA->z );

void vmp_QVecMagnitudeVU(float * const pfD, const vmp3DQVector * const pvA)
{
#ifdef USE_ASM
    sceVu0FVECTOR vD;

    ASSERT_PTR4(pfD);
    ASSERT_PTR16(pvA);

    asm __volatile__("

    lqc2    vf4,0x0(%1)         # A
        # Dot Product    
    vmul.xyz    vf5,vf4,vf4     # D.x = A.xyz * A.xyz
    vaddy.x     vf5,vf5,vf5     # D.x += D.y
    vaddz.x     vf5,vf5,vf5     # D.x += D.z
        # sqrt
    vsqrt Q,vf5x                # q=sqrt(Dx)
    vwaitq      
    vaddq.x vf4x, vf0x, Q       #  =0+q

    sqc2    vf4,0x0(%0)         # write float

    ":
     : "r" (vD) , "r" (pvA) 
    );

    *pfD = vD[0];
    
#else
    ASSERT_PTR4(pfD);
    ASSERT_PTR16(pvA);

    *pfD = sqrt( sceVu0InnerProduct( *(sceVu0FVECTOR *)pvA, *(sceVu0FVECTOR *)pvA) );
#endif
}



//______________________________________________________
//
//  (Quad) Vector Normalization - Unaligned
//
//  fMag = sqrtf( pvA->x * pvA->x + pvA->y * pvA->y + pvA->z * pvA->z );
//      fMag = 1.0f / fMag;
//      pvD->x = pvA->x * fMag;
//      pvD->y = pvA->y * fMag;
//      pvD->z = pvA->z * fMag;

void vmp_QVecNormalizeVU0(vmp3DQVector * const pvD, const vmp3DQVector * const pvA)
{
    sceVu0FVECTOR vD, vA;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);

    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
    vA[3] = 0.0f;

    sceVu0Normalize( vD, vA );
    pvD->x = vD[0];
    pvD->y = vD[1];
    pvD->z = vD[2];
    pvD->w = 0.0f;
}



//
//  (Quad) Vector Normalization - Aligned
//
//  fMag = sqrtf( pvA->x * pvA->x + pvA->y * pvA->y + pvA->z * pvA->z );
//      fMag = 1.0f / fMag;
//      pvD->x = pvA->x * fMag;
//      pvD->y = pvA->y * fMag;
//      pvD->z = pvA->z * fMag;

void vmp_QVecNormalizeVU(vmp3DQVector * const pvD, const vmp3DQVector * const pvA)
{
#if 0 //def USE_ASM
    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);

    asm __volatile__("

    lqc2    vf4,0x0(%1)         # A
        # Dot Product    
    vmul.xyz    vf5,vf4,vf4     # D.x = A.xyz * A.xyz
    vaddy.x     vf5,vf5,vf5     # D.x += D.y
    vaddz.x     vf5,vf5,vf5     # D.x += D.z
        # sqrt
    vsqrt Q,vf5x                # q=sqrt(Dx)
    vwaitq      
    vaddq.x vf5x, vf0x, Q       #  =0+q
        # fMag

#need to jump if 0 here!!!

    vdiv    Q, vf0w, vf5x       # Q = 1 / scalar
    vwaitq
    vmulq.xyzw  vf6,vf4,Q       # D.xyzw = A.xyzw / scalar
    
    sqc2    vf6,0x0(%0)         # write float

    ":
     : "r" (pvD) , "r" (pvA) 
    );
    
#else
    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);

    sceVu0Normalize( *(sceVu0FVECTOR *)pvD, *(sceVu0FVECTOR *)pvA );
#endif
}



//______________________________________________________
//
//  Setup Vec3D module
//
//  NOTE: This function sets up the function vectors so the best function
//      is attached to the vector for the processor running.
//

void QVsf3DSetup( const CpuInfo * const pcinfo )
{
    ASSERT_PTR(pcinfo);

    if ( CPUBITS_VU0 & pcinfo->nBits )
      {
        vmp_QVecSqrt =              vmp_QVecSqrtVU;
        vmp_QVecSqrt0 =             vmp_QVecSqrtVU0;
        vmp_QVecSqrtFast =          vmp_QVecSqrtVU;
        vmp_QVecSqrtFast0 =         vmp_QVecSqrtVU0;

        vmp_QVecMin =               vmp_QVecMinVU;
        vmp_QVecMin0 =              vmp_QVecMinVU0;
        vmp_QVecMax =               vmp_QVecMaxVU;
        vmp_QVecMax0 =              vmp_QVecMaxVU0;

        vmp_QVecMagnitude =         vmp_QVecMagnitudeVU;
        vmp_QVecMagnitude0 =        vmp_QVecMagnitudeVU0;
        vmp_QVecMagnitudeFast =     vmp_QVecMagnitudeVU;
        vmp_QVecMagnitudeFast0 =    vmp_QVecMagnitudeVU0;

        vmp_QVecNormalize =         vmp_QVecNormalizeVU;
        vmp_QVecNormalize0 =        vmp_QVecNormalizeVU0;
        vmp_QVecNormalizeFast =     vmp_QVecNormalizeVU;
        vmp_QVecNormalizeFast0 =    vmp_QVecNormalizeVU0;
      }

    else
      {
        vmp_QVecSqrt =              vmp_QVecSqrtGeneric;
        vmp_QVecSqrt0 =             vmp_QVecSqrtGeneric;
        vmp_QVecSqrtFast =          vmp_QVecSqrtGeneric;
        vmp_QVecSqrtFast0 =         vmp_QVecSqrtGeneric;

        vmp_QVecMin =               vmp_QVecMinGeneric;
        vmp_QVecMin0 =              vmp_QVecMinGeneric;
        vmp_QVecMax =               vmp_QVecMaxGeneric;
        vmp_QVecMax0 =              vmp_QVecMaxGeneric;

        vmp_QVecMagnitude =         vmp_QVecMagnitudeGeneric;
        vmp_QVecMagnitude0 =        vmp_QVecMagnitudeGeneric;
        vmp_QVecMagnitudeFast =     vmp_QVecMagnitudeGeneric;
        vmp_QVecMagnitudeFast0 =    vmp_QVecMagnitudeGeneric;

        vmp_QVecNormalize =         vmp_QVecNormalizeGeneric;
        vmp_QVecNormalize0 =        vmp_QVecNormalizeGeneric;
        vmp_QVecNormalizeFast =     vmp_QVecNormalizeGeneric;
        vmp_QVecNormalizeFast0 =    vmp_QVecNormalizeGeneric;
      }
}





