
//
//  Vsf3DVU.c - 3D Vector Special Function
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  NOTE: Because these a 3 float Vectors the 4th float has to be handled as 
//      a special case thus slowing the code down. This is the same as an AoS 
//      orientation. Instead a SoA should be used in conjunction with the quad
//      set of functions.
//
//  Created - 3 Oct 2001 - James Leiterman
//
//  Last Modification
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries
    
#include    <assert.h>
#include    <stdio.h>
#include    <math.h>

#ifdef CC_VMP_PS2
#include    <libvu0.h>
#endif

    //  Book Development Library Modules

#include    "CpuAsm.h"                                  // CPU
#include    "vmp3D.h"
#include    "vmp3DEx.h"

#include    "Vsf3D.h"                                   // 3D Math
#include    "Vsf3DEx.h"



//
//  Vector Square Root - Unaligned
//
//  pvD->x = sqrtf( pvA->x );
//  pvD->y = sqrtf( pvA->y );
//  pvD->z = sqrtf( pvA->z );
    
void vmp_VecSqrtVU0(vmp3DVector * const pvD, const vmp3DVector * const pvA)
{
    sceVu0FVECTOR vA, vD;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);

    ASSERT_FNEG(pvA->x);
    ASSERT_FNEG(pvA->y);
    ASSERT_FNEG(pvA->z);

    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;

    asm __volatile__("
    lqc2    vf5,0x0(%1)         #pvA

    vsqrt Q,vf5x                # q=sqrt(Ax)
    vwaitq      
    vaddq.x vf4x, vf0x, Q       #  =0+q

    vsqrt Q,vf5y                # q=sqrt(Ay)
    vwaitq      
    vaddq.y vf4y, vf0y, Q       #  =0+q
     
    vsqrt Q,vf5z                # q=sqrt(Az)
    vwaitq      
    vaddq.z vf4z, vf0z, Q       #  =0+q

    sqc2    vf4,0x0(%0)         # write Vector XYZ#
  
    ": : "r" (vD) , "r" (vA));

    pvD->x = vD[0];
    pvD->y = vD[1];
    pvD->z = vD[2];
}



//
//  Vector Square Root - Aligned
//
    
void vmp_VecSqrtVU(vmp3DVector * const pvD, const vmp3DVector * const pvA)
{
    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);

    ASSERT_FNEG(pvA->x);
    ASSERT_FNEG(pvA->y);
    ASSERT_FNEG(pvA->z);

    asm __volatile__("
    lqc2    vf4,0x0(%0)         #pvD .w
    lqc2    vf5,0x0(%1)         #pvA

    vsqrt Q,vf5x                # q=sqrt(Ax)
    vwaitq      
    vaddq.x vf4x, vf0x, Q       #  =0+q

    vsqrt Q,vf5y                # q=sqrt(Ay)
    vwaitq      
    vaddq.y vf4y, vf0y, Q       #  =0+q
     
    vsqrt Q,vf5z                # q=sqrt(Az)
    vwaitq      
    vaddq.z vf4z, vf0z, Q       #  =0+q

    sqc2    vf4,0x0(%0)         # write Vector XYZW
  
    ": : "r" (pvD) , "r" (pvA));
}



//______________________________________________________
//
//  Vector Minimum - Unaligned
//
//  pvD->x = (pvA->x < pvB->x) ? pvA->x : pvB->x;
//  pvD->y = (pvA->y < pvB->y) ? pvA->y : pvB->y;
//  pvD->z = (pvA->z < pvB->z) ? pvA->z : pvB->z;
    
void vmp_VecMinVU0( vmp3DVector * const pvD, const vmp3DVector * const pvA, const vmp3DVector * const pvB )
{
    sceVu0FVECTOR vA, vB, vD;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);
    ASSERT_PTR4(pvB);
 
    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;

    vB[0] = pvB->x;
    vB[1] = pvB->y;
    vB[2] = pvB->z;

    asm __volatile__("
    lqc2    vf6,0x0(%1)         #vA
    lqc2    vf5,0x0(%2)         #vB
    
    vmini.xyz vf4,vf6,vf5       # min XYZ

    sqc2    vf4,0x0(%0)         # write TMP Vector XYZW
  
    ": : "r" (vD) , "r" (vA), "r" (vB));

    pvD->x = vD[0];
    pvD->y = vD[1];
    pvD->z = vD[2];
}


//
//  Vector Minimum - Aligned
//
    
void vmp_VecMinVU( vmp3DVector * const pvD, const vmp3DVector * const pvA, const vmp3DVector * const pvB )
{
    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);
    ASSERT_PTR16(pvB);

    asm __volatile__("
    lqc2    vf4,0x0(%0)         #pvD
    lqc2    vf6,0x0(%1)         #pvA
    lqc2    vf5,0x0(%2)         #pvB
    
    vmove.w vf6,vf4             # copy .w field
    vmini.xyz vf4,vf6,vf5       # min XYZ

    sqc2    vf4,0x0(%0)         # write Vector XYZW
  
    ": : "r" (pvD) , "r" (pvA), "r" (pvB));
}




//______________________________________________________
//
//  Vector Maximum
//
//  pvD->x = (pvA->x > pvB->x) ? pvA->x : pvB->x;
//  pvD->y = (pvA->y > pvB->y) ? pvA->y : pvB->y;
//  pvD->z = (pvA->z > pvB->z) ? pvA->z : pvB->z;
    
void vmp_VecMaxVU0( vmp3DVector * const pvD, const vmp3DVector * const pvA, const vmp3DVector * const pvB )
{
    sceVu0FVECTOR vA, vB, vD;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);
    ASSERT_PTR4(pvB);
 
    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;

    vB[0] = pvB->x;
    vB[1] = pvB->y;
    vB[2] = pvB->z;

    asm __volatile__("
    lqc2    vf6,0x0(%1)         #vA
    lqc2    vf5,0x0(%2)         #vB
    
    vmax.xyz vf4,vf6,vf5        # max XYZ

    sqc2    vf4,0x0(%0)         # write TMP Vector XYZW
  
    ": : "r" (vD) , "r" (vA), "r" (vB));

    pvD->x = vD[0];
    pvD->y = vD[1];
    pvD->z = vD[2];
}



//
//  Vector Maximum - Aligned
//
    
void vmp_VecMaxVU( vmp3DVector * const pvD, const vmp3DVector * const pvA, const vmp3DVector * const pvB )
{
    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);
    ASSERT_PTR16(pvB);

    asm __volatile__("
    lqc2    vf4,0x0(%0)         #pvD
    lqc2    vf6,0x0(%1)         #pvA
    lqc2    vf5,0x0(%2)         #pvB
    
    vmove.w vf6,vf4             # copy .w field
    vmax.xyz vf4,vf6,vf5        # max XYZ

    sqc2    vf4,0x0(%0)         # write Vector XYZW
  
    ": : "r" (pvD) , "r" (pvA), "r" (pvB));
}



//______________________________________________________
//
//  Vector Magnitude - Unaligned
//
//  *pfD = sqrtf( pvA->x * pvA->x + pvA->y * pvA->y + pvA->z * pvA->z );

void vmp_VecMagnitudeVU0(float * const pfD, const vmp3DVector * const pvA)
{
    sceVu0FVECTOR vA;

    ASSERT_PTR4(pfD);
    ASSERT_PTR4(pvA);

    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
    vA[3] = 0.0f;
        
    *pfD = sqrt( sceVu0InnerProduct(vA, vA) );
}



//  Vector Magnitude - Aligned
//
//  *pfD = sqrtf( pvA->x * pvA->x + pvA->y * pvA->y + pvA->z * pvA->z );

void vmp_VecMagnitudeVU(float * const pfD, const vmp3DVector * const pvA)
{
    ASSERT_PTR4(pfD);
    ASSERT_PTR16(pvA);

    *pfD = sqrt( sceVu0InnerProduct( *(sceVu0FVECTOR *)pvA, *(sceVu0FVECTOR *)pvA) );
}



//______________________________________________________
//
//  Vector Normalization - Unaligned
//
//  fMag = sqrtf( pvA->x * pvA->x + pvA->y * pvA->y + pvA->z * pvA->z );
//      fMag = 1.0f / fMag;
//      pvD->x = pvA->x * fMag;
//      pvD->y = pvA->y * fMag;
//      pvD->z = pvA->z * fMag;

void vmp_VecNormalizeVU0(vmp3DVector * const pvD, const vmp3DVector * const pvA)
{
    sceVu0FVECTOR vD, vA;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);

    vA[0] = pvA->x;
    vA[1] = pvA->y;
    vA[2] = pvA->z;
    vA[3] = 0.0f;

    sceVu0Normalize( vD, vA );
    pvD->x = vD[0];
    pvD->y = vD[1];
    pvD->z = vD[2];
}



//
//  Vector Normalization - Aligned
//
//  fMag = sqrtf( pvA->x * pvA->x + pvA->y * pvA->y + pvA->z * pvA->z );
//      fMag = 1.0f / fMag;
//      pvD->x = pvA->x * fMag;
//      pvD->y = pvA->y * fMag;
//      pvD->z = pvA->z * fMag;

void vmp_VecNormalizeVU(vmp3DVector * const pvD, const vmp3DVector * const pvA)
{
    sceVu0FVECTOR vD;

    ASSERT_PTR16(pvD);
    ASSERT_PTR16(pvA);

    sceVu0Normalize( vD, *(sceVu0FVECTOR *)pvA );
        // Warning .W = 0.0
    pvD->x = vD[0];     // sceVu0CopyVectorXYZ()
    pvD->y = vD[1];
    pvD->z = vD[2];
}



//______________________________________________________
//
//  Setup Vec3D module
//
//  NOTE: This function sets up the function vectors so the best function
//      is attached to the vector for the processor running.
//

void Vsf3DSetup( const CpuInfo * const pcinfo )
{
    ASSERT_PTR4(pcinfo);

    if ( CPUBITS_VU0 & pcinfo->nBits )
      {
        vmp_VecSqrt =           vmp_VecSqrtVU;
        vmp_VecSqrt0 =          vmp_VecSqrtVU0;
        vmp_VecSqrtFast =       vmp_VecSqrtVU;
        vmp_VecSqrtFast0 =      vmp_VecSqrtVU0;

        vmp_VecMin =            vmp_VecMinVU;
        vmp_VecMin0 =           vmp_VecMinVU0;
        vmp_VecMax =            vmp_VecMaxVU;
        vmp_VecMax0 =           vmp_VecMaxVU0;

        vmp_VecMagnitude =      vmp_VecMagnitudeVU;
        vmp_VecMagnitude0 =     vmp_VecMagnitudeVU0;
        vmp_VecMagnitudeFast =  vmp_VecMagnitudeVU;
        vmp_VecMagnitudeFast0 = vmp_VecMagnitudeVU0;

        vmp_VecNormalize =      vmp_VecNormalizeVU;
        vmp_VecNormalize0 =     vmp_VecNormalizeVU0;
        vmp_VecNormalizeFast =  vmp_VecNormalizeVU;
        vmp_VecNormalizeFast0 = vmp_VecNormalizeVU0;
      }

    else
      {
        vmp_VecSqrt =           vmp_VecSqrtGeneric;
        vmp_VecSqrt0 =          vmp_VecSqrtGeneric;
        vmp_VecSqrtFast =       vmp_VecSqrtGeneric;
        vmp_VecSqrtFast0 =      vmp_VecSqrtGeneric;

        vmp_VecMin =            vmp_VecMinGeneric;
        vmp_VecMin0 =           vmp_VecMinGeneric;
        vmp_VecMax =            vmp_VecMaxGeneric;
        vmp_VecMax0 =           vmp_VecMaxGeneric;

        vmp_VecMagnitude =      vmp_VecMagnitudeGeneric;
        vmp_VecMagnitude0 =     vmp_VecMagnitudeGeneric;
        vmp_VecMagnitudeFast =  vmp_VecMagnitudeGeneric;
        vmp_VecMagnitudeFast0 = vmp_VecMagnitudeGeneric;

        vmp_VecNormalize =      vmp_VecNormalizeGeneric;
        vmp_VecNormalize0 =     vmp_VecNormalizeGeneric;
        vmp_VecNormalizeFast =  vmp_VecNormalizeGeneric;
        vmp_VecNormalizeFast0 = vmp_VecNormalizeGeneric;
      }
}





